import { isBuiltin } from 'node:module';
import micromatch from 'micromatch';
const IGNORE_DEFINITELY_TYPED = ['node'];
export default class DependencyDeputy {
    _manifests = new Map();
    manifests = new Map();
    canceledWorkspaces;
    referencedDependencies;
    peerDependencies;
    tsConfigPathGlobs = new Map();
    isProduction = false;
    constructor() {
        this.referencedDependencies = new Map();
        this.peerDependencies = new Map();
        this.canceledWorkspaces = new Set();
    }
    addWorkspace({ name, dir, manifestPath, manifest, isProduction, }) {
        const scripts = Object.values(manifest.scripts ?? {});
        const dependencies = Object.keys(manifest.dependencies ?? {});
        const peerDependencies = Object.keys(manifest.peerDependencies ?? {});
        const optionalDependencies = Object.keys(manifest.optionalDependencies ?? {});
        const devDependencies = Object.keys(manifest.devDependencies ?? {});
        const productionDependencies = [...dependencies, ...peerDependencies, ...optionalDependencies];
        this.manifests.set(name, manifest);
        this._manifests.set(name, {
            workspaceDir: dir,
            manifestPath,
            scripts,
            dependencies,
            peerDependencies,
            optionalDependencies,
            devDependencies,
            productionDependencies,
            allDependencies: [...productionDependencies, ...devDependencies],
        });
        this.isProduction = isProduction;
    }
    cancelWorkspace(workspaceName) {
        this.canceledWorkspaces.add(workspaceName);
    }
    getManifest(workspaceName) {
        return this.manifests.get(workspaceName);
    }
    getWorkspaceManifest(workspaceName) {
        return this._manifests.get(workspaceName);
    }
    getProductionDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.productionDependencies ?? [];
    }
    getDevDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.devDependencies ?? [];
    }
    getAllDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.allDependencies ?? [];
    }
    addTypeScriptConfigPathGlobs(workspaceName, paths) {
        this.tsConfigPathGlobs.set(workspaceName, paths ? Object.keys(paths).map(p => p.replace(/\*/g, '**')) : []);
    }
    addReferencedDependency(workspaceName, packageName) {
        if (!this.referencedDependencies.has(workspaceName)) {
            this.referencedDependencies.set(workspaceName, new Set());
        }
        this.referencedDependencies.get(workspaceName)?.add(packageName);
    }
    addPeerDependencies(workspaceName, peerDependencies) {
        this.peerDependencies.set(workspaceName, peerDependencies);
    }
    maybeAddListedReferencedDependency(workspace, moduleSpecifier, isStrict) {
        if (this.isInternalDependency(workspace.name, moduleSpecifier))
            return;
        const packageName = this.resolvePackageName(moduleSpecifier);
        const workspaceNames = isStrict ? [workspace.name] : [workspace.name, ...[...workspace.ancestors].reverse()];
        const closestWorkspaceName = workspaceNames.find(name => this.isInDependencies(name, packageName));
        if (closestWorkspaceName) {
            this.addReferencedDependency(closestWorkspaceName, packageName);
        }
        else {
            return moduleSpecifier;
        }
    }
    isInternalDependency(workspaceName, moduleSpecifier) {
        if (moduleSpecifier.startsWith('/') || moduleSpecifier.startsWith('.'))
            return true;
        if (isBuiltin(moduleSpecifier))
            return true;
        const packageName = this.resolvePackageName(moduleSpecifier);
        return (!this.isInDependencies(workspaceName, packageName) && this.isAliasedDependency(workspaceName, moduleSpecifier));
    }
    isAliasedDependency(workspaceName, moduleSpecifier) {
        const patterns = this.tsConfigPathGlobs.get(workspaceName);
        return patterns && patterns.length > 0 && micromatch.isMatch(moduleSpecifier, patterns);
    }
    resolvePackageName(moduleSpecifier) {
        const parts = moduleSpecifier.split('/').slice(0, 2);
        return moduleSpecifier.startsWith('@') ? parts.join('/') : parts[0];
    }
    isInDependencies(workspaceName, packageName) {
        const manifest = this._manifests.get(workspaceName);
        return Boolean(manifest?.allDependencies.includes(packageName));
    }
    settleDependencyIssues() {
        const dependencyIssues = [];
        const devDependencyIssues = [];
        for (const [workspaceName, { manifestPath }] of this._manifests.entries()) {
            if (this.canceledWorkspaces.has(workspaceName))
                continue;
            const referencedDependencies = this.referencedDependencies.get(workspaceName);
            const isUnreferencedDependency = (dependency) => {
                const [scope, typedDependency] = dependency.split('/');
                if (scope === '@types') {
                    if (IGNORE_DEFINITELY_TYPED.includes(typedDependency))
                        return false;
                    if (referencedDependencies?.has(dependency))
                        return false;
                    const [scope, name] = typedDependency.split('__');
                    const typedPackageName = scope && name ? `@${scope}/${name}` : typedDependency;
                    return referencedDependencies ? !referencedDependencies.has(typedPackageName) : false;
                }
                if (!referencedDependencies?.has(dependency)) {
                    const peerDependencies = Array.from(this.peerDependencies.get(workspaceName)?.get(dependency) ?? []);
                    return !peerDependencies.find(peerDependency => !isUnreferencedDependency(peerDependency));
                }
                return false;
            };
            this.getProductionDependencies(workspaceName)
                .filter(isUnreferencedDependency)
                .forEach(symbol => dependencyIssues.push({ filePath: manifestPath, symbol }));
            this.getDevDependencies(workspaceName)
                .filter(isUnreferencedDependency)
                .forEach(symbol => devDependencyIssues.push({ filePath: manifestPath, symbol }));
        }
        return { dependencyIssues, devDependencyIssues };
    }
}
