import { ts, Node } from 'ts-morph';
import { _findExternalImportModuleSpecifiers } from './util/externalImports.js';
import { findUnusedClassMembers, findUnusedEnumMembers } from './util/members.js';
import { _findDuplicateExportedNames, _hasReferencingDefaultImport, _findReferencingNamespaceNodes, _getExportedDeclarations, _findReferences, hasExternalReferences, } from './util/project.js';
import { getType } from './util/type.js';
export default class SourceLab {
    issueManager;
    report;
    workspaceDirs;
    skipExportsAnalysis;
    isReportExports;
    isReportValues;
    isReportTypes;
    constructor({ issueManager, report, workspaceDirs }) {
        this.issueManager = issueManager;
        this.report = report;
        this.workspaceDirs = workspaceDirs;
        this.skipExportsAnalysis = new Set();
        this.isReportValues = report.exports || report.nsExports || report.classMembers;
        this.isReportTypes = report.types || report.nsTypes || report.enumMembers;
        this.isReportExports = this.isReportValues || this.isReportTypes;
    }
    skipExportsAnalysisFor(filePath) {
        if (typeof filePath === 'string')
            this.skipExportsAnalysis.add(filePath);
        else
            filePath.forEach(filePath => this.skipExportsAnalysis.add(filePath));
    }
    analyzeSourceFile(sourceFile) {
        const im = this.issueManager;
        const report = this.report;
        let externalModuleSpecifiers = [];
        im.counters.processed++;
        const filePath = sourceFile.getFilePath();
        if (report.dependencies || report.unlisted) {
            externalModuleSpecifiers = _findExternalImportModuleSpecifiers(sourceFile);
        }
        if (report.duplicates) {
            const duplicateExports = _findDuplicateExportedNames(sourceFile);
            duplicateExports.forEach(symbols => {
                const symbol = symbols.join('|');
                im.addIssue('duplicates', { filePath, symbol, symbols });
            });
        }
        if (this.skipExportsAnalysis.has(filePath))
            return externalModuleSpecifiers;
        if (this.isReportExports) {
            this.analyzeExports(sourceFile);
        }
        return externalModuleSpecifiers;
    }
    analyzeExports(sourceFile) {
        const im = this.issueManager;
        const report = this.report;
        const filePath = sourceFile.getFilePath();
        const exportDeclarations = _getExportedDeclarations(sourceFile);
        exportDeclarations.forEach(declarations => {
            declarations.forEach(declaration => {
                const type = getType(declaration);
                if (type && !this.isReportTypes)
                    return;
                if (!type && !this.isReportValues)
                    return;
                if (ts.getJSDocPublicTag(declaration.compilerNode))
                    return;
                if (Node.isJSDocTag(declaration))
                    return;
                let identifier;
                let fakeIdentifier;
                if (declaration.isKind(ts.SyntaxKind.EnumDeclaration)) {
                    identifier = declaration.getFirstChildByKind(ts.SyntaxKind.Identifier);
                    if (report.enumMembers) {
                        findUnusedEnumMembers(declaration, filePath).forEach(member => im.addIssue('enumMembers', {
                            filePath,
                            symbol: member.getName(),
                            parentSymbol: identifier?.getText(),
                        }));
                    }
                }
                else if (declaration.isKind(ts.SyntaxKind.ClassDeclaration)) {
                    identifier = declaration.getFirstChildByKind(ts.SyntaxKind.Identifier);
                    if (report.classMembers) {
                        findUnusedClassMembers(declaration, filePath).forEach(member => im.addIssue('classMembers', {
                            filePath,
                            symbol: member.getName(),
                            parentSymbol: identifier?.getText(),
                        }));
                    }
                }
                if (Node.isExportGetable(declaration) &&
                    declaration.isDefaultExport() &&
                    _hasReferencingDefaultImport(sourceFile)) {
                    return;
                }
                if (!identifier) {
                    if (declaration.isKind(ts.SyntaxKind.Identifier)) {
                        identifier = declaration;
                    }
                    else if (declaration.isKind(ts.SyntaxKind.ArrowFunction) ||
                        declaration.isKind(ts.SyntaxKind.ObjectLiteralExpression) ||
                        declaration.isKind(ts.SyntaxKind.ArrayLiteralExpression) ||
                        declaration.isKind(ts.SyntaxKind.StringLiteral) ||
                        declaration.isKind(ts.SyntaxKind.NumericLiteral)) {
                        if (!_hasReferencingDefaultImport(sourceFile)) {
                            fakeIdentifier = 'default';
                        }
                    }
                    else if (declaration.isKind(ts.SyntaxKind.FunctionDeclaration) ||
                        declaration.isKind(ts.SyntaxKind.TypeAliasDeclaration) ||
                        declaration.isKind(ts.SyntaxKind.InterfaceDeclaration)) {
                        identifier = declaration.getFirstChildByKind(ts.SyntaxKind.Identifier);
                    }
                    else if (declaration.isKind(ts.SyntaxKind.PropertyAccessExpression)) {
                        identifier = declaration.getLastChildByKind(ts.SyntaxKind.Identifier);
                    }
                    else {
                        identifier = declaration.getFirstDescendantByKind(ts.SyntaxKind.Identifier);
                    }
                }
                if (identifier || fakeIdentifier) {
                    const identifierText = fakeIdentifier ?? identifier?.getText() ?? '*';
                    if (report.exports && im.issues.exports[filePath]?.[identifierText])
                        return;
                    if (report.types && im.issues.types[filePath]?.[identifierText])
                        return;
                    if (report.nsExports && im.issues.nsExports[filePath]?.[identifierText])
                        return;
                    if (report.nsTypes && im.issues.nsTypes[filePath]?.[identifierText])
                        return;
                    const refs = _findReferences(identifier);
                    if (refs.length === 0) {
                        im.addIssue('exports', { filePath, symbol: identifierText });
                    }
                    else {
                        if (hasExternalReferences(refs, filePath))
                            return;
                        if (_findReferencingNamespaceNodes(sourceFile).length > 0) {
                            if (type) {
                                im.addIssue('nsTypes', { filePath, symbol: identifierText, symbolType: type });
                            }
                            else {
                                im.addIssue('nsExports', { filePath, symbol: identifierText });
                            }
                        }
                        else if (type) {
                            im.addIssue('types', { filePath, symbol: identifierText, symbolType: type });
                        }
                        else {
                            im.addIssue('exports', { filePath, symbol: identifierText });
                        }
                    }
                }
            });
        });
    }
}
