import path from 'node:path';
import * as npm from './npm-scripts/index.js';
import * as plugins from './plugins/index.js';
import { ROOT_WORKSPACE_NAME, TEST_FILE_PATTERNS } from './util/constants.js';
import { debugLogFiles, debugLogIssues } from './util/debug.js';
import { _pureGlob, negate } from './util/glob.js';
const negatedTestFilePatterns = TEST_FILE_PATTERNS.map(negate);
export default class WorkspaceWorker {
    name;
    dir;
    config;
    ancestorManifests;
    rootWorkspaceConfig;
    rootConfig;
    referencedDependencyIssues = new Set();
    manifest;
    rootWorkspaceDir;
    referencedDependencies = new Set();
    peerDependencies = new Map();
    negatedWorkspacePatterns = [];
    enabled;
    isRoot;
    constructor({ name, dir, config, ancestorManifests, rootWorkspaceConfig, rootConfig, negatedWorkspacePatterns, manifest, rootWorkspaceDir, }) {
        this.name = name;
        this.dir = dir;
        this.config = config;
        this.ancestorManifests = ancestorManifests;
        this.isRoot = name === ROOT_WORKSPACE_NAME;
        this.rootConfig = rootConfig;
        this.rootWorkspaceConfig = rootWorkspaceConfig;
        this.rootWorkspaceDir = rootWorkspaceDir;
        this.negatedWorkspacePatterns = negatedWorkspacePatterns;
        this.manifest = manifest;
        this.enabled = {
            babel: false,
            capacitor: false,
            changesets: false,
            cypress: false,
            eslint: false,
            gatsby: false,
            jest: false,
            next: false,
            nx: false,
            playwright: false,
            postcss: false,
            remark: false,
            remix: false,
            rollup: false,
            storybook: false,
        };
    }
    getConfigForPlugin(pluginName) {
        return (this.config[pluginName] ??
            this.rootWorkspaceConfig[pluginName] ?? { config: [], entryFiles: [], projectFiles: [] });
    }
    async init() {
        this.setEnabledPlugins();
        await this.initReferencedDependencies();
    }
    setEnabledPlugins() {
        const dependencies = new Set([this.manifest, ...this.ancestorManifests]
            .flatMap(manifest => [Object.keys(manifest.dependencies ?? {}), Object.keys(manifest.devDependencies ?? {})])
            .flat());
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            const hasIsEnabled = typeof plugin.isEnabled === 'function';
            this.enabled[pluginName] = hasIsEnabled && plugin.isEnabled({ manifest: this.manifest, dependencies });
        }
    }
    async initReferencedDependencies() {
        const { dependencies, peerDependencies } = await npm.findDependencies(this.rootConfig.ignoreBinaries, this.manifest, this.isRoot, this.dir, this.rootWorkspaceDir);
        const filePath = path.join(this.dir, 'package.json');
        dependencies.forEach(dependency => this.referencedDependencyIssues.add({ filePath, symbol: dependency }));
        dependencies.forEach(dependency => this.referencedDependencies.add(dependency));
        this.peerDependencies = peerDependencies;
    }
    getEntryFilePatterns() {
        const { entryFiles } = this.config;
        if (entryFiles.length === 0)
            return [];
        return [entryFiles, TEST_FILE_PATTERNS, this.isRoot ? this.negatedWorkspacePatterns : []]
            .flat()
            .map(p => p.replace(/!$/, ''));
    }
    getProjectFilePatterns() {
        const { projectFiles } = this.config;
        if (projectFiles.length === 0)
            return [];
        return [projectFiles, TEST_FILE_PATTERNS, this.isRoot ? this.negatedWorkspacePatterns : []]
            .flat()
            .map(p => p.replace(/!$/, ''));
    }
    getPluginEntryFilePatterns(isProduction = false) {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabled[pluginName]) {
                const pluginConfig = this.getConfigForPlugin(pluginName);
                const defaultConfigPatterns = 'ENTRY_FILE_PATTERNS' in plugin ? plugin.ENTRY_FILE_PATTERNS : [];
                patterns.push(...defaultConfigPatterns);
                patterns.push(...pluginConfig.entryFiles);
                if (!isProduction) {
                    const defaultConfigPatterns = 'PRODUCTION_ENTRY_FILE_PATTERNS' in plugin ? plugin.PRODUCTION_ENTRY_FILE_PATTERNS : [];
                    patterns.push(...defaultConfigPatterns);
                }
            }
        }
        return patterns.flat().map(p => p.replace(/!$/, ''));
    }
    getPluginProjectFilePatterns() {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabled[pluginName]) {
                const pluginConfig = this.getConfigForPlugin(pluginName);
                const entryFilesPatterns = pluginConfig.entryFiles;
                const projectFilePatterns = pluginConfig.projectFiles;
                const defaultEntryFilePatterns = 'ENTRY_FILE_PATTERNS' in plugin ? plugin.ENTRY_FILE_PATTERNS : [];
                const defaultProjectFilePatterns = 'PROJECT_FILE_PATTERNS' in plugin ? plugin.PROJECT_FILE_PATTERNS : defaultEntryFilePatterns;
                patterns.push(defaultProjectFilePatterns);
                patterns.push(projectFilePatterns.length > 0 ? projectFilePatterns : entryFilesPatterns);
            }
        }
        return patterns.flat().map(p => p.replace(/!$/, ''));
    }
    getPluginConfigPatterns() {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabled[pluginName]) {
                const pluginConfig = this.getConfigForPlugin(pluginName);
                const defaultConfigPatterns = 'CONFIG_FILE_PATTERNS' in plugin ? plugin.CONFIG_FILE_PATTERNS : [];
                patterns.push(...defaultConfigPatterns);
                patterns.push(...pluginConfig.config);
            }
        }
        return patterns.flat().map(p => p.replace(/!$/, ''));
    }
    getProductionEntryFilePatterns() {
        const entryFiles = this.config.entryFiles.filter(p => p.endsWith('!'));
        if (entryFiles.length === 0)
            return [];
        const negatedEntryFiles = this.config.entryFiles.filter(p => !p.endsWith('!')).map(negate);
        return [entryFiles, negatedEntryFiles, negatedTestFilePatterns, this.isRoot ? this.negatedWorkspacePatterns : []]
            .flat()
            .map(p => p.replace(/!$/, ''));
    }
    getProductionProjectFilePatterns() {
        const projectFiles = this.config.projectFiles;
        if (projectFiles.length === 0)
            return this.getProductionEntryFilePatterns();
        const _projectFiles = this.config.projectFiles.map(p => {
            if (!p.endsWith('!') && !p.startsWith('!'))
                return negate(p);
            return p;
        });
        const negatedEntryFiles = this.config.entryFiles.filter(p => !p.endsWith('!')).map(negate);
        const negatedPluginConfigPatterns = this.getPluginConfigPatterns().map(negate);
        const negatedPluginEntryFilePatterns = this.getPluginEntryFilePatterns(true).map(negate);
        const negatedPluginProjectFilePatterns = this.getPluginProjectFilePatterns().map(negate);
        return [
            _projectFiles,
            negatedEntryFiles,
            negatedPluginConfigPatterns,
            negatedPluginEntryFilePatterns,
            negatedPluginProjectFilePatterns,
            negatedTestFilePatterns,
            this.isRoot ? this.negatedWorkspacePatterns : [],
        ]
            .flat()
            .map(p => p.replace(/!$/, ''));
    }
    getProductionPluginEntryFilePatterns() {
        const patterns = [];
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabled[pluginName]) {
                if ('PRODUCTION_ENTRY_FILE_PATTERNS' in plugin) {
                    const pluginConfig = this.getConfigForPlugin(pluginName);
                    const defaultConfigPatterns = plugin.PRODUCTION_ENTRY_FILE_PATTERNS;
                    patterns.push(...defaultConfigPatterns);
                    patterns.push(...pluginConfig.entryFiles);
                }
            }
        }
        const p = patterns.flat().map(p => p.replace(/!$/, ''));
        if (p.length === 0)
            return [];
        return [p, negatedTestFilePatterns].flat();
    }
    getConfigurationEntryFilePattern(pluginName) {
        const pluginConfig = this.getConfigForPlugin(pluginName);
        return [...plugins[pluginName].CONFIG_FILE_PATTERNS, ...pluginConfig.config];
    }
    getWorkspaceIgnorePatterns() {
        const ignoreFiles = this.rootConfig.ignoreFiles;
        const ignore = this.config.ignore;
        return [...ignoreFiles, ...ignore];
    }
    async findDependenciesByPlugins() {
        for (const [pluginName, plugin] of Object.entries(plugins)) {
            if (this.enabled[pluginName]) {
                const hasDependencyFinder = 'findDependencies' in plugin && typeof plugin.findDependencies === 'function';
                if (hasDependencyFinder) {
                    const dependencies = await this.findDependenciesByPlugin(pluginName, plugin.findDependencies);
                    dependencies.forEach(dependency => this.referencedDependencyIssues.add(dependency));
                    dependencies.forEach(dependency => this.referencedDependencies.add(dependency.symbol));
                }
            }
        }
        return {
            referencedDependencyIssues: this.referencedDependencyIssues,
            referencedDependencies: this.referencedDependencies,
        };
    }
    async findDependenciesByPlugin(pluginName, pluginCallback) {
        const patterns = this.getConfigurationEntryFilePattern(pluginName);
        const cwd = this.dir;
        const ignore = this.getWorkspaceIgnorePatterns();
        const configFilePaths = await _pureGlob({ patterns, cwd, ignore });
        debugLogFiles(1, `Globbed ${pluginName} config file paths`, configFilePaths);
        if (configFilePaths.length === 0)
            return [];
        const referencedDependencies = (await Promise.all(configFilePaths.map(async (configFilePath) => {
            const dependencies = await pluginCallback(configFilePath, { cwd, manifest: this.manifest });
            return dependencies.map(symbol => ({ filePath: configFilePath, symbol }));
        }))).flat();
        debugLogIssues(1, `Dependencies used by ${pluginName} configuration`, referencedDependencies);
        return referencedDependencies;
    }
}
